
// Alex Sherman (asherman@cs.columbia.edu) 
// 2009 
// The AConfig class takes care of reading and setting configuration options
// from "conf" and "conf.allowed" files. 


package alex;

import java.io.*;
import java.util.*;
import org.gudy.azureus2.core3.config.COConfigurationManager;



public class AConfig {
    // DEFAULTS:

    // We make the default connections to be 250/500. This is set in 
    // this way to be comparable to PropShare and BitTyrant - when comparing
    // against those clients. If comparing against plain Azureus we should
    // these to 50 and 250 respectively for fair comparison. 
    public int DefaultConnectionsPerTorrent = 250;
    public int DefaultTotalConnections = 500;

    // Dafault relative path to where the file is saved.
    public String DefaultSavePath = "my_dir";
    
    public String FairTorrentClientType = "FairTorrent-1.1.1"; 
    // Azureus 3.0.4.2 is the code base uses for this implementation of 
    // FairTorrent. 
    public String AzureusClientType = "Azureus-3.0.4.2";

    public int DefaultPort = 7081; // Azureus default port is: 6881

    public int DefaultLogLevel = 0; 

    // PARAMETERS: 

    //if true: run in fairtorrent mode
    public boolean fairtorrent_mode;

    //if true then runs seeds in fairtorrent mode also 
    public boolean seed_fairtorrent_mode;

    // type of client
    public String my_client_type;

    public int connections_per_torrent;
    public int total_connections;

    // Default max upload rate in KBytes/s. If -1, then unlimited.
    // i.e. determined by the system. Default: -1.
    public int max_upload_rate;

    // Default max download rate in KBytes/s. If -1, then unlimited, 
    // i.e. determined by the system. Default: -1.
    public int max_download_rate;
    
    public int my_port; 
    public int command_port;

    public int log_level; 
    
    // the relative path where the file is saved 
    public String save_path;
    
    // max number of unchokes for leecher and seed. This is not applicable
    // for FairTorrent as FairTorrent unchokes all eligible leecher 
    // neighbors that have interest in the data. 
    // For non-FairTorrent mode if -1 (default is used), then allows
    // the system to determin the number of unchoked conns. (Typically
    // several connections). Otherwise
    // if a positive integer sets the number of unchokes to that
    // integer.
    public int num_unchokes_leecher; 
    public int num_unchokes_seed; 

    // WE DO NOT RECOMMEND PLAYING WITH THE REST OF THE PARAMETERS
    // IN THIS CLASS: 

    //if true: allows connections only to/from ips in allowed conf 
    //(i.e. if true, treat allowed conf as a white list) 
    //default: false - allow ips not listed in allowed conf 
    public boolean only_ips_in_allowed_conf;

    //if true: does not allow connections to/from ips in allowed conf 
    //(i.e if true, treat allowd conf as a black list) 
    //default: false - allow ips listed in allowed conf 
    public boolean not_ips_in_allowed_conf;
    
    //if true: I act as free rider and don't honor requests for data 
    public boolean am_free_rider;

    //if true: don't upload when finished downloading 
    //this would apply to a seed as well, therefore it's better 
    //to configure this for leechers only.
    public boolean dont_seed_when_done;
    
    //if true: means I finished downloading
    public boolean done_with_download;

    //if true (and fairtorrent_mode), then uses ft_stack_opt optimization
    public boolean ft_stack_opt; 

    //if ft_stack optimization is used, then this is the max size of the stack
    public int ft_stack_size; 

    //if > 0 and fairtorrent_mode is true, then uses this as the message size
    //in fairtorrent
    public int fairtorrent_msg_size; 

    //When X it will cause the burst upload rate to be at most the number
    //of (bytes allowed to be sent per second) * (X)  
    public double burst_rate_ratio; 
    
    //number of seconds to sleep before start 
    public int sleep_time; 

    //if not a seed, and departure_time > 0, then 
    //exit the system "departure_time" seconds after completing download
    public int departure_time; 

    private Map amap; 
    
    public AConfig() {

	amap = new HashMap();

	// set defauls, read conf files, change global Azureus settings:
	configure();

    } 
 
    class allowed_peer {
	public String ctype; 
    } 

    public void read_conf(String filename) {
	
	String line; 
	//ArrayList lines = new ArrayList();
	try
	    {    
		BufferedReader in = new BufferedReader(new FileReader(filename));
		
		if (!in.ready()) { 
		    throw new IOException();
		}


		ALogger.log(0,"...");
		ALogger.log(0,"Reading Conf Values: ");

		while ((line = in.readLine()) != null) { 
		    //lines.add(line);
		    
		    String[] commented = line.split("#");
		    //System.out.println(line);
		    String[] words = commented[0].split(" +"); 
		    int len = words.length;
		    if (len !=2) {
			continue; 
		    } 

		    if (words[0].equals("fairtorrent_mode")) {
			if(Integer.parseInt(words[1]) == 0) {
			    fairtorrent_mode = false; 
			} 
		    } 
		    
		    if (words[0].equals("seed_fairtorrent_mode")) {
			if(Integer.parseInt(words[1]) == 0) {
			    seed_fairtorrent_mode = false; 
			} 
		    } 

		    if (words[0].equals("conns")) {
			if(Integer.parseInt(words[1]) > -1) { 
			    connections_per_torrent = Integer.parseInt(words[1]);
			}
		    } 

		    if (words[0].equals("total_conns")) {
			if(Integer.parseInt(words[1]) > -1) { 
			    total_connections = Integer.parseInt(words[1]);
			}
		    } 
		    
		    if (words[0].equals("max_upload_rate")) {
			max_upload_rate = Integer.parseInt(words[1]); 
		    } 
		    if (words[0].equals("max_download_rate")) {
			max_download_rate = Integer.parseInt(words[1]); 
			ALogger.log(0,"max_download_rate " + max_download_rate);
		    } 

		    if (words[0].equals("port")) {
			command_port = Integer.parseInt(words[1]);
			my_port = command_port +1;
		    }

		    if (words[0].equals("log_level")) {
			log_level = Integer.parseInt(words[1]);
		    }
 
		    if (words[0].equals("save_path")) {
			    save_path = words[1]; 
		    } 

		    if (words[0].equals("num_unchokes_leecher")) {
			if(Integer.parseInt(words[1]) >= -1) { 
			    num_unchokes_leecher = Integer.parseInt(words[1]);
			}
		    } 

		    if (words[0].equals("num_unchokes_seed")) {
			if(Integer.parseInt(words[1]) >= -1) { 
			    num_unchokes_seed = Integer.parseInt(words[1]);
			}
		    } 
		    
		    // we keep these options configurable - but these options
		    // are not applicable to the public realease and should
		    // not be listed in the conf file. 
		    if (words[0].equals("ft_stack_opt")) {
			if(Integer.parseInt(words[1]) == 1) {
			    ft_stack_opt = true; 
			} 
		    } 
		    
		    if (words[0].equals("ft_stack_size")) {
			if(Integer.parseInt(words[1]) > 0) {
			    ft_stack_size =Integer.parseInt(words[1]); 
			}
		    } 
		    
		    if (words[0].equals("only_ips_in_allowed_conf")) {
			if(Integer.parseInt(words[1]) != 0) {
			    only_ips_in_allowed_conf = true; 
			} 
			ALogger.log(0,"only_ips_in_allowed_conf " + only_ips_in_allowed_conf);
		    } 
		    if (words[0].equals("not_ips_in_allowed_conf")) {
			if(Integer.parseInt(words[1]) != 0) {
			    not_ips_in_allowed_conf = true; 
			} 
			ALogger.log(0,"not_ips_in_allowed_conf " + not_ips_in_allowed_conf);
		    } 
		    if (words[0].equals("dont_seed_when_done")) {
			if(Integer.parseInt(words[1]) != 0) {
			    dont_seed_when_done = true; 
			} 
			ALogger.log(0,"dont_seed_when_done " + dont_seed_when_done);
		    } 
		    if (words[0].equals("sleep")) {
			if(Integer.parseInt(words[1]) > 0) {
			    sleep_time =  Integer.parseInt(words[1]);
			} 
		    } 
		    
		    if (words[0].equals("departure_time")) {
			if(Integer.parseInt(words[1]) > 0) {
			    departure_time =  Integer.parseInt(words[1]);
			} 
		    }
 
		    
		}
		
		in.close();
			
		if (sleep_time > 0) { 
		    try { 
			ALogger.log(0,"sleeping " + sleep_time + " seconds");
			Thread.sleep(sleep_time*1000);
		    }
		    catch (InterruptedException e) {
			e.printStackTrace();
		    } 
		}
	    }
        catch (IOException e)
	    {
		System.out.println(e);
		return;
	    }	
	
    } 


    public void read_conf_allowed(String allowedFile) {
	String line;
	try
	    {    
		BufferedReader in = new BufferedReader(new FileReader(allowedFile));
		
		if (!in.ready()) { 
		    throw new IOException();
		}


		ALogger.log(0,"...");
		ALogger.log(0,"Reading Conf Allowed: ");

		while ((line = in.readLine()) != null) { 
		    //lines.add(line);
		    
		    String[] commented = line.split("#");
		    //System.out.println(line);
		    String[] words = commented[0].split(" +"); 
		    int len = words.length;
		    if (len <2) {
			continue; 
		    } 

		    String ip = words[1];
		    allowed_peer ap = new allowed_peer();
		    amap.put(ip, ap);
		    
		    for(int i=0; i<len; i++) {
			String[] pair = words[i].split("=");
			if (pair[0].equals("class")) {
			    ap.ctype = pair[1];
			} 
		    } 
		}

		
		in.close();
	    }
        catch (IOException e)
	    {
		System.out.println(e);
		return;
	    }	
	
	Iterator it = amap.entrySet().iterator();

	while(it.hasNext()){
	    Map.Entry entry = (Map.Entry)it.next();
	    String ip = (String)entry.getKey();
            allowed_peer ap = (allowed_peer)entry.getValue();
	    
	    ALogger.log(0, "Key " + ip + " Value " + ap.ctype);
	}
    } 

    public int getLogLevel() {
	return log_level;
    } 


    //gets class type of IP or returns "-";
    public String get_ctype(String ip) {
	if (!amap.containsKey(ip)) {
	    return "-"; 
	} 
	else { 
	    allowed_peer ap = (allowed_peer)amap.get(ip);
	    return ap.ctype;
	}
    } 

    public String getMyClientType() {
	return my_client_type; 
    } 

    public void setMyClientType(String s) {
	my_client_type = s; 
    } 


    //when download is completed this could be called by AStats 
    public void download_done() {
	if (dont_seed_when_done) {
	    am_free_rider = true; 
	} 
	done_with_download = true;
    } 

    //can I connect to the given IP ? 
    public boolean containsAllowedIp(String ip) {
	if (only_ips_in_allowed_conf) { 
	    return amap.containsKey(ip);
	}
	else if (not_ips_in_allowed_conf) {
	    return !amap.containsKey(ip);
	} 
	else { 
	    return true;
	} 
    } 

    public void set_defaults() {
	// By default fairtorrent mode is on: 
	fairtorrent_mode = true; 
	seed_fairtorrent_mode = true; 
	my_client_type = FairTorrentClientType;	
	
	connections_per_torrent = DefaultConnectionsPerTorrent;
	total_connections = DefaultTotalConnections;
	
	max_upload_rate = -1;
	max_download_rate = -1;
	
	my_port = DefaultPort;
	command_port = DefaultPort-1;
	log_level = DefaultLogLevel;
	save_path = DefaultSavePath;
	
	num_unchokes_leecher = -1;
	num_unchokes_seed = -1;
	
	// We don't recommend changing the defaults below. These variables
	// were created for various type of experimentation and should not
	// be applicable to most users.
	only_ips_in_allowed_conf = false;
	not_ips_in_allowed_conf = false; 
	am_free_rider = false;  
	dont_seed_when_done = false; 
	done_with_download = false; 
	
	fairtorrent_msg_size = 17000;
	
	//ft stack optimizaion: 
	ft_stack_opt = false;
	ft_stack_size = 100; 
	
	burst_rate_ratio = 1.2; 
	
	sleep_time = -1;
	departure_time = -1;
    }
    
    private void configure() { 

	// 1. Set default values for params
	set_defaults();

	// 1A. read the values from the config file that may override
	// the default settings.
	read_conf("conf");
	
	// 2. conf.allowed
	// conf.allowed is used to list IPs that we can talk to. This is
	// important for experimental purpose - to make sure that we only
	// talk to a configured set of IPs. This is irrelevant for the 
	// public release of the code. 
	read_conf_allowed("conf.allowed");
	
	// Now use the values specified in the conf, which may override the
	// defaults set the Azureus parameters based on these values. 

	// 3. set the client type
	if (!fairtorrent_mode) {
	    my_client_type = AzureusClientType;
	}
	
	// 4. if upload/download rates were listed in the conf set them here:
	if (max_upload_rate > -1) { 
	    
	    if (max_upload_rate == 0) {
		max_upload_rate = 1; //just in case set it to 1KB/s
		am_free_rider = true; //but we also mark it as free rider 
	    } 
	    COConfigurationManager.setParameter("Max Upload Speed KBs", max_upload_rate);
	    COConfigurationManager.setParameter("Max Upload Speed Seeding KBs", max_upload_rate);
	    //does not really matter b/c we are disabling LAN Speed below...
	    //COConfigurationManager.setParameter("Max LAN Upload Speed KBs", 1);
	    
	} 
	if (max_download_rate > -1) { 
	    COConfigurationManager.setParameter("Max Download Speed KBs", max_download_rate);
	    COConfigurationManager.setParameter("Max LAN Download Speed KBs", max_download_rate);
	} 

	//by default LAN speed is scheduled differently by Azureus 
	//and throws off measurements in a lab that includes machines
	//on the same LAN. We disable the LAN speed configuration
	//in order not to throw off those measurements.
	COConfigurationManager.setParameter("LAN Speed Enabled", 0);

	//5. configure ports 
	if (my_port > -1) {
	    COConfigurationManager.setParameter("TCP.Listen.Port", my_port);
	    COConfigurationManager.setParameter("UDP.Listen.Port", my_port);
	    COConfigurationManager.setParameter("UDP.NonData.Listen.Port", my_port);
	} 

	// 6. Configure number of connections per torrent and global based
	// on the defaults and/or values from conf.
	reconfigure_connections();

	// 7 configure num unchokes (Azureus refers to these as "Max Uploads"
	//in BitTorrent terminology this is the number of unchoked connections

	//If the fairtorrent_mode is on, then the number of unchokes
	//is irrelevant. FairTorrent unchokes all of its leecher neighbors.
	//See the code of "fairtorrentCalculateUnchokes" in these files:
	//"com/aelitis/azureus/core/peermanager/unchoker/{DownloadingUnchoker.java,SeedingUnchoker.java}"
	//If the fairtorrent_mode is off, by default we rely on the default
	//behavior of Azureus which unchokes several neighbors at a time.
	//However if the fairtorrent_mode is on, and num_unchokes_leecher
	//is specified, the we set the max unchokes to num_unchokes_leecher.
	//Same logic applies to unchokes by the seeders. 
	if (!fairtorrent_mode && num_unchokes_leecher > 0) {
	    COConfigurationManager.setParameter("Max Uploads", 
						num_unchokes_leecher);
	}
	if (!seed_fairtorrent_mode && num_unchokes_seed > 0) {
	    COConfigurationManager.setParameter("Max Uploads Seeding", 
						num_unchokes_seed);
	}


	// 8 some other defaults...
	COConfigurationManager.setParameter("Enable incremental file creation", 1);
	COConfigurationManager.setParameter("diskmanager.perf.cache.enable.read", 1);
	COConfigurationManager.setParameter("diskmanager.perf.cache.size", 16);

	//path where the file is saved 
	if (save_path != "") { 
	    COConfigurationManager.setParameter("Default save path", save_path);
	}

	// 9. Print out default settings:
	ALogger.log(0,"fairtorrent_mode:\t" + fairtorrent_mode);
	ALogger.log(0,"seed_fairtorrent_mode:\t" + seed_fairtorrent_mode);
	ALogger.log(0,"port:\t" + my_port);
	ALogger.log(0,"connections_per_torrent:\t" + connections_per_torrent);
	ALogger.log(0,"num_unchokes_leecher:\t" + num_unchokes_leecher);
	ALogger.log(0,"num_unchokes_seed:\t" + num_unchokes_seed);
	ALogger.log(0,"total_connections:\t" + total_connections);
	ALogger.log(0,"max_upload_rate: (KBytes/s)\t" + max_upload_rate);
	ALogger.log(0,"max_download_rate: (KBytes/s)\t" + max_download_rate);
	ALogger.log(0,"save_path:\t" + save_path);
	ALogger.log(0,"log_level:\t" + log_level + " Default log_level: 0\t");
	
	ALogger.log(0,"\n\nClient STARTED:\t" + my_client_type + "\n\n");
    }

    public void reconfigure_connections() {
	if (connections_per_torrent == 0) {
	    connections_per_torrent = DefaultConnectionsPerTorrent;
	}
	if (total_connections < connections_per_torrent) {
	    total_connections = connections_per_torrent;
	} 
	
	COConfigurationManager.setParameter("Max.Peer.Connections.Per.Torrent", 
					    connections_per_torrent); 
	COConfigurationManager.setParameter("Max.Peer.Connections.Total", 
					    total_connections);
	
	// This is not necessary in FT/Azureus, but just to make a 
	// distinction with BitTyrant that sets this value to TRUE
	COConfigurationManager.setParameter("Enable.AutoMaxConnections", 0);
    }
	
}



