
// Alex Sherman (asherman@cs.columbia.edu) 
// 2009 
// The AStats class takes care of keeping track of statistics, especially
// of the bytes uploaded to and downloaded from each peer. 

package alex;

import java.io.*;
import java.util.*;
import org.gudy.azureus2.core3.util.*; //SystemTime.java comes from here, also timer stuff 

import org.gudy.azureus2.core3.disk.*; //DiskManager and stuff... 
import org.gudy.azureus2.core3.download.*; //DownloadManager
import org.gudy.azureus2.core3.global.*; //GlobalManager
import org.gudy.azureus2.core3.tracker.client.*; //TRTrackerAnnouncer
import org.gudy.azureus2.core3.config.COConfigurationManager;

import org.gudy.azureus2.core3.peer.impl.transport.PEPeerTransportProtocol;
import com.aelitis.azureus.core.peermanager.utils.*; //OutgoingBTPieceMessageHandler
import com.aelitis.azureus.core.peermanager.messaging.*;
import com.aelitis.azureus.core.networkmanager.*; //NetworkConnection


public class AStats implements TimerEventPerformer {
    
    private double completion_time; 
    private int stats_log_period = 15; //15 seconds 
    private int updates_period = 6; //6 seconds 

    private long last_stats_time = 0; 
    private double pdone = 0.0; //PercentDone 
    
    private DiskManager disk_mgr; 
    private GlobalManager glob_mgr = null; 
    
    private boolean initialized = false; 
    private boolean dm_initialized = false; 
    
    public Map pmap; //map of peers 

    private DownloadManager download_manager;

    private long last_time_tracker_announce = 0; 
    private int tracker_announce_interval = 30; //30 seconds 
    private double announce_after_percent = 0.95; //perform regular tracker announce after this percent of the file is downloaded 

    private long protocol_bytes_sent = 0; 
    private long protocol_bytes_received = 0; 
    
    private Random randgen = new Random(SystemTime.getCurrentTime());
    private int max_token = 1000; 

    //Max positive and negative deficits 
    public int max_pos_deficit = 0; 
    public int max_neg_deficit = 0; 
    public int cur_deficit = 0; 
    
    private final AEMonitor updates_mon  = new AEMonitor( "AStats:updates" );
    private int loop_count = 0; 

    private ArrayList updates_lst  = new ArrayList(); 
    private ArrayList old_updates_lst  = new ArrayList(); 
    
    private final AEMonitor ft_stack_lock = new AEMonitor( "AStats:fts_lock" );
    private LinkedList ft_stack = new LinkedList();

    private final int BYTES_RECEIVED = 0; 
    private final int BYTES_SENT = 1; 
    

    public int high_priority_conns = 0;  
    public int normal_priority_conns = 0;  

    public long time_completed = 0; //SystemTime.getCurrentTime();// 0; 

    public long time_to_depart = 0; //time to depart the system 

    AStats() {
	pmap = new HashMap();
	SimpleTimer.addPeriodicEvent("AStats", 1000,this);
    } 

    private int get_rand_token() {
	return (int)(randgen.nextDouble() * (double)max_token);
    } 

    public String get_ctype(String ip) {
	return AManager.getConfObject().get_ctype(ip);
    } 

    public void initialize(DiskManager mgr) {
	disk_mgr = mgr; 
	initialized = true;
	ALogger.log(0, "AStats initialized");
	//Reconfigure connections one more time -- to be sure:
	AManager.getConfObject().reconfigure_connections();
    } 

    public void initDownloadManager(DownloadManager mgr) {
	download_manager = mgr; 
	dm_initialized = true;
	ALogger.log(ALogger.LOG_DEBUG, "AStats initialized download manager");
    } 

    public apeer getPeerById(byte []id) {
	if (!pmap.containsKey(id)) {
	    return null;
	}
	apeer ap = (apeer)pmap.get(id); 
	return ap;
    } 

    public int getNumHighPriorityConnections() {
	return high_priority_conns; 
    } 

    public void setNumHighPriorityConnections(int n) {
	high_priority_conns = n;
    } 

    public int getNumNormalPriorityConnections() {
	return normal_priority_conns; 
    } 

    public void setNumNormalPriorityConnections(int n) {
	normal_priority_conns = n;
    } 

    public void setGlobalManager(GlobalManager m) {
	glob_mgr = m;
    } 

    //if in fairtorrent mode and ft stack optimization is on,
    //then if we have owe data to this id, and we have messages or requests
    //to fill to that guy push him as the last on the stack 
    public void pieceReceivedFrom(byte []id, String ip, NetworkConnection conn, OutgoingBTPieceMessageHandler outPHandler) {
	if (AManager.getConfObject().fairtorrent_mode &&
	    AManager.getConfObject().ft_stack_opt) {
	    
	    if (conn == null || outPHandler == null) {
		return; 
	    } 
	    
	    NetworkConnectionBase connb = (NetworkConnectionBase)conn;
	    
	    if (getDeficit(id) < 0 &&
		(connb.getOutgoingMessageQueue().alexGetNumDataMsgs() > 0 ||
		 outPHandler.getRequestCount() > 0)) { 

		try { 
		    ft_stack_lock.enter();
		    ft_stack.addLast(new aPeerConn(id, conn, ip));
		    ALogger.log(0, "FTSTACK pushing " + ip + " on the stack. Stack size: " + ft_stack.size());
		    if (ft_stack.size() > AManager.getConfObject().ft_stack_size) {
			ft_stack.removeFirst();
		    } 
		} 
		catch( Throwable e ) { 
		    ALogger.log(0, "FTSTACK error in pieceReceivedFrom");
		}
		finally {
		    ft_stack_lock.exit();
		} 
	    }
	} 
	
    } 

    //pop the first conn from the stack to whom we owe data 
    public aPeerConn popConnFromFtOptStack() {
	aPeerConn apc = null; 
	
	if (AManager.getConfObject().fairtorrent_mode &&
	    AManager.getConfObject().ft_stack_opt) {
	    try { 
		ft_stack_lock.enter();
		if(ft_stack.size() > 0) { 
		    apc = (aPeerConn)ft_stack.removeFirst();
		    ALogger.log(0, "FTSTACK popping " + apc.ip + " from the stack. Stack size: " + ft_stack.size());
		} 
	    } 
	    catch( Throwable e ) { 
		ALogger.log(0, "FTSTACK error in popConnFromFtOptStack");
	    }
	    finally {
		ft_stack_lock.exit();
	    } 
	}
	
	return apc;
    } 

    public void dataBytesSent(PEPeerTransportProtocol transp, byte []id, String ip, int bytes, boolean bSeed) {
	
	
	if (!pmap.containsKey(id)) {
	    pmap.put(id, new apeer(id, ip, get_ctype(ip), get_rand_token()));
	} 
	//ALogger.log(0, "DEFICIT-ACCNT STORE UPLOAD: " + ip + " " + bytes);
	store_data(transp, id, bytes, BYTES_SENT); 

	apeer ap = (apeer)pmap.get(id); 
	ap.total_uploaded += bytes;
	ap.last_sent = SystemTime.getCurrentTime();
	ap.update(bSeed);
    }
    
    public void _data_sent(PEPeerTransportProtocol transp, byte []id, int bytes) {

	try { 
	    
	    apeer ap = (apeer)pmap.get(id); 


	    //ALogger.log(0, "DEFICIT-ACCNT PROC UPLOAD: " + ap.peer_ip + " " + bytes + " seed " + transp.am_seed);
		
	    if (AManager.getConfObject().done_with_download || 
		transp.am_seed) {
		return;
	    } 
	    
	    cur_deficit += bytes; 
	    if (cur_deficit > max_pos_deficit) {
		max_pos_deficit = cur_deficit; 
	    } 
	    
	}catch (Exception e) {
	    //e.printStackTrace();
	    ALogger.log(0, "EXCEPTION caught on _data_sent");
	} 

    } 

    public void dataBytesReceived(PEPeerTransportProtocol transp, byte []id, String ip, int bytes, boolean bSeed) {
	if (!pmap.containsKey(id)) {
	    pmap.put(id, new apeer(id, ip, get_ctype(ip), get_rand_token()));
	} 
	
	//ALogger.log(0, "DEFICIT-ACCNT STORE DOWNLOAD: " + ip + " " + bytes);
	store_data(transp, id, bytes, BYTES_RECEIVED); 

	apeer ap = (apeer)pmap.get(id);
	ap.total_downloaded += bytes;
	ap.last_received = SystemTime.getCurrentTime();
	ap.update(bSeed); 
    }
    public void _data_received(PEPeerTransportProtocol transp, byte []id, int bytes) {

	try { 
	    
	    apeer ap = (apeer)pmap.get(id);

	    //ALogger.log(0, "DEFICIT-ACCNT PROC DOWNLOAD: " + ap.peer_ip + " " + bytes + " seed " + transp.am_seed);

	    if (AManager.getConfObject().done_with_download || 
		transp.am_seed) {
		return;
	    } 

	    cur_deficit -= bytes; 
	    if (-cur_deficit > max_neg_deficit) {
		max_neg_deficit = -cur_deficit; 
	    } 
	}catch (Exception e) {
	    //e.printStackTrace();
	    ALogger.log(0, "EXCEPTION caught on _data_received");
	} 
	
    }

    public void store_data(PEPeerTransportProtocol transp, byte []id,  int bytes, int flag) {
	updates_mon.enter();
	updates_lst.add(new aDataReport(transp, id, bytes, flag));
	updates_mon.exit();
    } 

    public void process_stored_data() {
	
	//ALogger.log(0, "ACCNT process_stored_data called"); 

	Iterator it = old_updates_lst.iterator();
	while(it.hasNext()) {
	    aDataReport r = (aDataReport)it.next();
	    if (r._flag == BYTES_RECEIVED) {
		_data_received(r._transp, r._id, r._bytes);
	    } 
	    else {
		_data_sent(r._transp, r._id, r._bytes);
	    } 
	} 

	updates_mon.enter(); 

	old_updates_lst = updates_lst;
	updates_lst.clear();
	
	updates_mon.exit();
       
    } 

    public void protocolBytesSent(byte []id, String ip, int bytes, boolean bSeed) {
	protocol_bytes_sent += bytes;
    } 

    public void protocolBytesReceived(byte []id, String ip, int bytes, boolean bSeed) {
	protocol_bytes_received += bytes; 
    }


    public int getDeficit(byte []id) {
	if (!pmap.containsKey(id)) {
	    return 0;
	} 
	
	apeer ap = (apeer)pmap.get(id);
	return ap.get_deficit();
    } 

    public int getRandomToken(byte []id) {
	if (!pmap.containsKey(id)) {
	    return max_token;
	} 
	
	apeer ap = (apeer)pmap.get(id);
	return ap.get_random_token();
    } 


    public void downloadCompleted(long tm) {
	completion_time = ((double)tm)/1000.0; 
	if (completion_time <0) {
	    completion_time = 0.0; 
	} 
	AManager.getConfObject().download_done();

	time_completed = SystemTime.getCurrentTime();
	
	
	ALogger.log(ALogger.DEBUG, "COMPLETED DOWNLOAD of file no-name time: " + completion_time + " bytes: none-given");

	if (AManager.getConfObject().departure_time > 0 && 
	    !AManager.getConfObject().dont_seed_when_done) {
	
	    time_to_depart = time_completed + (AManager.getConfObject().departure_time*1000);

	    ALogger.log(0, "will shut down in " + AManager.getConfObject().departure_time + " seconds");
	}

    } 

    public void  perform(TimerEvent event) {


	loop_count ++; 

	long now =  SystemTime.getCurrentTime();
	
	if (time_completed > 0 && 
	    time_to_depart > 0) { 

	    //if there is some time before we depart shoot off some 
	    //tracker requests - to report our progress 
	    if (now < time_to_depart - 10*1000 && 
		loop_count % 10 == 0) { 
		
		try { 
		    long tm = SystemTime.getCurrentTime();
		    if (dm_initialized) { 
			download_manager.requestTrackerAnnounce(true);
			
			ALogger.log(0, "TRACKER REQUEST INITIATED"); 
			
			
		    } 
		}catch (Exception e) {
		    e.printStackTrace();
		    ALogger.log(0, "EXCEPTION caught on download_manager access");
		} 
	    }
		
	    if (now >= time_to_depart) { 
		ALogger.log(0, "Shutting down");
		if (glob_mgr != null) {
		    glob_mgr.stopAllDownloads();
		    glob_mgr.stopGlobalManager();
		} 
		
		Runtime r = Runtime.getRuntime();
		try { 
		    r.exec("mkdir logs"); 
		    String done_file = "done.txt";
		    r.exec("touch " + done_file);
		    PrintWriter pw = new PrintWriter(new FileWriter(done_file, true));
		    pw.println("DONE");
		}
		catch (IOException e) {
		    e.printStackTrace();
		} 
		
		System.exit(0);
	    }
	} 

		

	if (loop_count % updates_period == 0) {
	    process_stored_data(); 
	} 
	if (loop_count % stats_log_period != 0) {
	    return; 
	} 

	
	log_stats();
	
	int maxUpSpeed             = COConfigurationManager.getIntParameter("Max Upload Speed KBs",0);
	int maxDownSpeed    = COConfigurationManager.getIntParameter("Max Download Speed KBs",0);
	
	ALogger.log(ALogger.DEBUG, "CONF up speed: " + maxUpSpeed + " down speed: " + maxDownSpeed);
    }

    private void log_stats() {

	try { 
	    if (initialized) {
		pdone = ((double)disk_mgr.getPercentDone())/100.0;
		//ALogger.log(0, "GETTING PERCENT " + disk_mgr.getPercentDone());
		//ALogger.log(0, "Total Pieces " + disk_mgr.getNbPieces() + " remains: " + disk_mgr.getRemainingExcludingDND());
		
		final long total =disk_mgr.getTotalLength();
		
		
		int my_completion = 0; 
		if (total > 0) { 
		    my_completion = 
			(int) ((1000 * (total - disk_mgr.getRemainingExcludingDND())) / total);
		}
		pdone = ((double)my_completion)/1000.0;
		
	    } 
	}catch (Exception e) {
	    e.printStackTrace();
	    ALogger.log(0, "EXCEPTION caught on disk_manager access");
	} 

	try { 
	    long tm = SystemTime.getCurrentTime();
	    if (dm_initialized && 
		((tm -last_time_tracker_announce)/1000 >= 
		 tracker_announce_interval) && 
		pdone > announce_after_percent) {
		
		TRTrackerAnnouncer tc = download_manager.getTrackerClient();
		if (tc!=null) { 
		    tc.update(true);
		}
		last_time_tracker_announce = tm; 
		
	    } 
	}catch (Exception e) {
	    e.printStackTrace();
	    ALogger.log(0, "EXCEPTION caught on download_manager access");
	} 

	String sdone = "Percent Done: " + pdone; 
	ALogger.log(0,sdone); 
	    
	long cur_time = SystemTime.getCurrentTime();
	
	
	int utotal_leech = 0; 
	int dtotal_leech = 0; 
	int dtotal_seed = 0; 
	
	int udiff = 0; 
	int leechd_diff = 0; 
	int seedd_diff = 0; 
	
	int total_deficit = 0; 
	
	Iterator it = pmap.entrySet().iterator();

	while(it.hasNext()){
	    Map.Entry entry = (Map.Entry)it.next();
            apeer ap = (apeer)entry.getValue();
	    
	    
	    utotal_leech += ap.total_uploaded; 
	    if (!ap.am_seed) {
		dtotal_leech += ap.total_downloaded;  
	    } 
	    else {
		dtotal_leech += ap.leech_total_downloaded; 
		dtotal_seed += (ap.total_downloaded - ap.leech_total_downloaded);
	    } 

	    
	    ap.tmp_udiff = (ap.total_uploaded - ap.old_total_uploaded);
	    udiff += ap.tmp_udiff; 

	    ap.tmp_leechd_diff = 0; 
	    ap.tmp_seedd_diff = 0; 
	    
	    if (!ap.am_seed) {
		ap.tmp_leechd_diff =  (ap.total_downloaded - ap.old_total_downloaded); 
	    } 
	    else {
		if (ap.became_seed > last_stats_time) {
		    
		    ap.tmp_leechd_diff = (ap.leech_total_downloaded - ap.old_total_downloaded); 
		    ap.tmp_seedd_diff = (ap.total_downloaded - ap.leech_total_downloaded);
		} 
		else { 
		    ap.tmp_seedd_diff = (ap.total_downloaded - ap.old_total_downloaded);

		}
	    } 

	    leechd_diff += ap.tmp_leechd_diff; 
	    seedd_diff += ap.tmp_seedd_diff; 
	    
	}

		
	total_deficit = utotal_leech - dtotal_leech;

	double time_diff = (cur_time - last_stats_time)/1000.0;
	double leechDrate = (leechd_diff/1024.0)/time_diff; 
	double leechUrate = (udiff/1024.0)/time_diff;
	double seedDrate = (seedd_diff/1024.0)/time_diff;
	
	leechDrate = (int)(leechDrate*100)/100.0; 
	leechUrate = (int)(leechUrate*100)/100.0; 
	seedDrate = (int)(seedDrate*100)/100.0; 

	String s = "Total BSTAT " + leechDrate + " " + leechUrate + " " + 
	    seedDrate + " " + time_diff + " (ldbytes, lubytes, sdbytes, time)"; 
	ALogger.log(0,s); 
	
	s = "Max deficit(pos,neg,cur) (" + max_pos_deficit + "," + max_neg_deficit + "," + cur_deficit + ")";
	ALogger.log(0,s);

	s = "Total Deficit " + total_deficit + " ldbytes " + dtotal_leech + " tdbytes " + (dtotal_leech + dtotal_seed) + " protou " + protocol_bytes_sent + " protod " + protocol_bytes_received; 
	ALogger.log(0,s); 
	
	
	List idList = new ArrayList();

	Iterator it2 = pmap.entrySet().iterator();

	while(it2.hasNext()){
	    Map.Entry entry = (Map.Entry)it2.next();
            apeer ap = (apeer)entry.getValue();

	    idList.add(ap.peer_id);
		    
	    ap.save_old();

	}

	
	Object[] arr = idList.toArray();
	Arrays.sort(arr, new ap_compare());

	int len = arr.length;
	for(int i=0; i<len; i++) {
	    apeer ap = (apeer)pmap.get(arr[i]);
	    
	    int flag = ap.am_seed ? 1:0;
	    
	    s = "DSTATS " + ap.get_deficit() + " " + ap.tmp_udiff + " " + (ap.tmp_leechd_diff + ap.tmp_seedd_diff) + " " + ap.total_downloaded + " " + ap.ctype + " " + ap.peer_ip; 
	    s = s + " S:" + flag + " (x,x) (deficit, delta_u, delta_d, dtotal)";

	    ALogger.log(0,s);
		    
	} 

	last_stats_time = cur_time;
    }

}


class ap_compare implements Comparator {
    public final int compare (Object id1, Object id2) {
	
	apeer ap1 = (apeer)AManager.getStatsObject().pmap.get(id1);
	apeer ap2 = (apeer)AManager.getStatsObject().pmap.get(id2);
	
	if (!ap1.am_seed && ap2.am_seed) {
	    return -1;
	} 

	else if (ap1.am_seed && !ap2.am_seed) {
	    return 1; 
	} 

	else {
	    int def1 = ap1.get_deficit();
	    int def2 = ap2.get_deficit();
	    if (def1 > def2) {
		return 1;
	    } 
	    else if (def2 > def1) {
		return -1; 
	    } 
	    else {
		return 0; 
	    } 
	} 
    }
} 

