/*
 * Created on Sep 28, 2004
 * Created by Alon Rohter
 * Copyright (C) 2004, 2005, 2006 Aelitis, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * AELITIS, SAS au capital de 46,603.30 euros
 * 8 Allee Lenotre, La Grille Royale, 78600 Le Mesnil le Roi, France.
 *
 */

package com.aelitis.azureus.core.networkmanager.impl;

import java.io.IOException;
import java.util.*;

import org.gudy.azureus2.core3.util.*;

import com.aelitis.azureus.core.networkmanager.*;
import com.aelitis.azureus.core.peermanager.messaging.*;

//ALEX
import alex.*;
import java.net.InetSocketAddress;

/**
 * A rate-controlled write entity backed by multiple peer connections, with an
 * emphasis on transmitting packets with full payloads, i.e. it writes to the
 * transport in mss-sized chunks if at all possible. It also employs fair,
 * round-robin write scheduling, where connections each take turns writing a
 * single full packet per round.
 */
public class MultiPeerUploader implements RateControlledEntity {
  private static final int FLUSH_CHECK_LOOP_TIME = 500;  //500ms
  private static final int FLUSH_WAIT_TIME = 3*1000;  //3sec no-new-data wait before forcing write flush
  private long last_flush_check_time = 0;
  
  
  private final RateHandler rate_handler;
  private boolean destroyed = false;
  
  private final HashMap waiting_connections = new HashMap();
  private final LinkedList ready_connections = new LinkedList();
  private final AEMonitor lists_lock = new AEMonitor( "PacketFillingMultiPeerUploader:lists_lock" );
  
    //ALEX FT vars:
    //index of the current connection 
    private int cur_conn_index = 0;
    private int last_bytes_written = 1; //if 0 causes extra wait 
    private long last_time_written = 0; 

    //ALEX_DEBUG
    private long last_check =   SystemTime.getCurrentTime();
    private long last_partial_write_check = SystemTime.getCurrentTime();

  /**
   * Create a new packet-filling multi-peer upload entity,
   * rate-controlled by the given handler.
   * @param rate_handler listener to handle upload rate limits
   */
  public MultiPeerUploader( RateHandler rate_handler ) {
    this.rate_handler = rate_handler;
  }
  
  
  /**
   * Checks the connections in the waiting list to see if it's time to be force-flushed.
   */
  private void flushCheck() {
    long diff = SystemTime.getCurrentTime() - last_flush_check_time;
    
    if( !destroyed && (diff > FLUSH_CHECK_LOOP_TIME || diff < 0 ) ) {
      try {  lists_lock.enter();
        long current_time = SystemTime.getCurrentTime();
        
        for( Iterator i = waiting_connections.entrySet().iterator(); i.hasNext(); ) {
          Map.Entry entry = (Map.Entry)i.next();
          PeerData peer_data = (PeerData)entry.getValue();
          
          long wait_time = current_time - peer_data.last_message_added_time;
          
          if( wait_time > FLUSH_WAIT_TIME || wait_time < 0 ) {  //time to force flush
        	  
        	NetworkConnectionBase conn = (NetworkConnectionBase)entry.getKey();
            
            if( conn.getOutgoingMessageQueue().getTotalSize() > 0 ) { //has data to flush
              conn.getOutgoingMessageQueue().cancelQueueListener( peer_data.queue_listener ); //cancel the listener
              i.remove();  //remove from the waiting list
              addToReadyList( conn );
            }
            else { //no data, so reset flush wait time
            	
           	  peer_data.last_message_added_time = current_time;
            }
          }
        }
      }
      finally {  lists_lock.exit();  }
      
      last_flush_check_time = SystemTime.getCurrentTime();
    }
  }
  
  
  
  /**
   * Destroy this upload entity.
   * Note: Removes all peer connections in the process.
   */
  public void destroy() {
    destroyed = true;
    
    try {
      lists_lock.enter();
      
      //remove and cancel all connections in waiting list    
      for( Iterator i = waiting_connections.entrySet().iterator(); i.hasNext(); ) {
        Map.Entry entry = (Map.Entry)i.next();
        NetworkConnectionBase conn = (NetworkConnectionBase)entry.getKey();
        PeerData data = (PeerData)entry.getValue();
        conn.getOutgoingMessageQueue().cancelQueueListener( data.queue_listener );
      }
      waiting_connections.clear();
      
      //remove from ready list
      ready_connections.clear();
    }
    finally {
      lists_lock.exit();
    }
  }
  

  
  
  /**
   * Add the given connection to be managed by this upload entity.
   * @param peer_connection to be write managed
   */
  public void addPeerConnection( NetworkConnectionBase peer_connection ) {
    int mss_size = peer_connection.getMssSize();
    boolean has_urgent_data = peer_connection.getOutgoingMessageQueue().hasUrgentMessage();
    int num_bytes_ready = peer_connection.getOutgoingMessageQueue().getTotalSize();
    
    if( num_bytes_ready >= mss_size || has_urgent_data ) {  //has a full packet's worth, or has urgent data
      addToReadyList( peer_connection );
    }
    else {   //has data to send, but not enough for a full packet
      addToWaitingList( peer_connection );
    }
  }
  
  
  /**
   * Remove the given connection from this upload entity.
   * @param peer_connection to be removed
   * @return true if the connection was found and removed, false if not removed
   */
  public boolean removePeerConnection( NetworkConnectionBase peer_connection ) {
    try {
      lists_lock.enter();
      
      //look for the connection in the waiting list and cancel listener if found
      PeerData peer_data = (PeerData)waiting_connections.remove( peer_connection );
      if( peer_data != null ) {
        peer_connection.getOutgoingMessageQueue().cancelQueueListener( peer_data.queue_listener );
        return true;
      }
      
      //look for the connection in the ready list
      if( ready_connections.remove( peer_connection ) ) {
        return true;
      }
      
      return false;
    }
    finally {
      lists_lock.exit();
    }
  }
  
  
  

  //connections with less than a packet's worth of data
  private void addToWaitingList( final NetworkConnectionBase conn ) {
    final PeerData peer_data = new PeerData();
    
    OutgoingMessageQueue.MessageQueueListener listener = new OutgoingMessageQueue.MessageQueueListener() {
      public boolean messageAdded( Message message ) {  return true;  }
      
      public void messageQueued( Message message ) {  //connection now has more data to send
        try {
          lists_lock.enter();
          
          if( waiting_connections.get( conn ) == null ) {  //connection has already been removed from the waiting list
            return;  //stop further processing
          }
          
          int mss_size = conn.getMssSize();
          boolean has_urgent_data = conn.getOutgoingMessageQueue().hasUrgentMessage();
          int num_bytes_ready = conn.getOutgoingMessageQueue().getTotalSize();
        
          if( num_bytes_ready >= mss_size || has_urgent_data ) {  //has a full packet's worth, or has urgent data
            waiting_connections.remove( conn );  //remove from waiting list
            conn.getOutgoingMessageQueue().cancelQueueListener( this ); //cancel this listener
            addToReadyList( conn );
          }
          else {  //still not enough data for a full packet
        	if ( !peer_data.bumped ){
        		
               	// only do this once to avoid the possibility that, for example, sending a have
            	// every 2.9 seconds can result in them being delayed for a significant amount
            	// of time (e.g. 60 seconds +...)

        		peer_data.bumped = true;
        		peer_data.last_message_added_time = SystemTime.getCurrentTime();  //update last message added time
        	}
          }
        }
        finally {
          lists_lock.exit();
        }
      }

      public void messageRemoved( Message message ) {/*ignore*/}
      public void messageSent( Message message ) {/*ignore*/}
      public void protocolBytesSent( int byte_count ) {/*ignore*/}
      public void dataBytesSent( int byte_count ) {/*ignore*/}
    };
    
    peer_data.queue_listener = listener;  //attach listener
    peer_data.last_message_added_time = SystemTime.getCurrentTime(); //start flush wait time
    peer_data.bumped = false;
    
    try {
      lists_lock.enter();
      
      waiting_connections.put( conn, peer_data ); //add to waiting list
      conn.getOutgoingMessageQueue().registerQueueListener( listener );  //listen for added data
    }
    finally {
      lists_lock.exit();
    }
  }
  
  
  
  //connections ready to write
  private void addToReadyList( final NetworkConnectionBase conn ) {
    try {
      lists_lock.enter();
      
      ready_connections.addLast( conn );  //add to ready list
    }
    finally {
      lists_lock.exit();
    }
  }
  
  

    //ALEX: implementing fairtorrent_write
    private int fairtorrent_write( EventWaiter waiter, int num_bytes_to_write, boolean seed_mode) {
	HashMap connections_to_notify_of_exception = new HashMap();
	ArrayList manual_notifications = new ArrayList();
	ArrayList leftover_conns = new ArrayList();

	int num_bytes_remaining = num_bytes_to_write;    
	long now = SystemTime.getCurrentTime();
	
	if (last_bytes_written ==0) {
	    if (now - last_time_written < 50) { //same value as IDLE_SLEE_TIME
		//ALogger.log(0, "FT: fairtorrent_write called too early"); 
		return 0; 
	    } 
	} 

	boolean check_partial_write = false; 
	if (now - last_partial_write_check > 500) {
	    check_partial_write = true;
	    last_partial_write_check = now; 
	} 


	//ALogger.log(0, "FT: fairtorrent_write called"); 


	try {
	    lists_lock.enter();

	    //now loop through the list and send any control messages queued 
	    //in front of data messages 
	    for (Iterator i = ready_connections.iterator(); 
		 //we go on if either we are not checking for partial writes, or we have 
		 //bytes remaining (otherwise if we are doing partial writes of data 
		 //and no bytes remain - we stop) 
		 i.hasNext() && (!check_partial_write || num_bytes_remaining > 0);) { 
		
		NetworkConnectionBase conn = (NetworkConnectionBase)i.next();
		
		if (!conn.getTransportBase().isReadyForWrite(waiter)) {
		    continue;
		} 
		

		try {
		    //get the size of all protocol messages queued in front of the first data message 
		    int to_write = conn.getOutgoingMessageQueue().alexGetProtocolBytes();
		    int written  = 0; 
		    

		    int partial = 0; 

		    if (to_write <=0 && 
			check_partial_write) {
			
			//if no protocol msg, and it's time for a partial message check 
			//check if any partial message is to be written: 
			to_write = conn.getOutgoingMessageQueue().alexGetRemainingDataBytes();

			int mss_size = conn.getMssSize();
			int use_size = mss_size; 
			if (AManager.getConfObject().fairtorrent_msg_size > 0) {
			    use_size = AManager.getConfObject().fairtorrent_msg_size;
			} 

			to_write = use_size > to_write ? to_write : use_size; 
			partial = to_write; 
		    } 
		    
		    
		    if (to_write > 0) { 
			//write what can be written 
			written = conn.getOutgoingMessageQueue().deliverToTransport(to_write, true);
		    } 
		    
		    if( written > 0 ) {  
			manual_notifications.add( conn );  //register it for manual listener notification
			num_bytes_remaining -= written; 
			

			if (true) { 
			    InetSocketAddress notional_address = conn.getEndpoint().getNotionalAddress();
			    
			    String ip    = notional_address.getAddress().getHostAddress();
			    if (partial > 0 && 
				written > 0) {
				ALogger.log(ALogger.LOG_DEBUG2, "PARTIAL " + written + " bytes to " + ip);
			    }  
				
			    //ALogger.log(0, "FT PROTO: writing " + written + " bytes to " + ip);
			} 
			
		    }
		} 
		catch( Throwable e ) {  //write exception, so move to waiting list while it waits for removal
		    
		    if( AEDiagnostics.TRACE_CONNECTION_DROPS ) {
			if( e.getMessage() == null ) {
			    Debug.out( "null write exception message: ", e );
			}
			else {
			    if( e.getMessage().indexOf( "An existing connection was forcibly closed by the remote host" ) == -1 &&
				e.getMessage().indexOf( "Connection reset by peer" ) == -1 &&
				e.getMessage().indexOf( "Broken pipe" ) == -1 &&
				e.getMessage().indexOf( "An established connection was aborted by the software in your host machine" ) == -1 ) {
				
				System.out.println( "MP: write exception [" +conn.getTransportBase().getDescription()+ "]: " +e.getMessage() );
			    }
			}
		    }
		    
		    if (! (e instanceof IOException )){
			
			Debug.printStackTrace(e);
		    }
		    
		    connections_to_notify_of_exception.put( conn, e );  //do exception notification outside of sync'd block
		    //alex: not waiting list necessary 
		    //addToWaitingList( conn );
		}
	    } 
	    
	    //NetworkConnectionBase ready_array[] =  (NetworkConnectionBase[])ready_connections.toArray();
	    Object ready_array[] =  ready_connections.toArray();

	    int num_ready_connections = ready_connections.size();
	    //ALogger.log(0, "FT sz(ready_connection) = " + num_ready_connections + " bytes_remain = " + num_bytes_remaining);
	    while( num_bytes_remaining > 0 && num_ready_connections > 0) {
		
		num_ready_connections = ready_connections.size();
		
		//ALogger.log(-1, "FT sz(ready_connection) = " + num_ready_connections + " bytes_remain = " + num_bytes_remaining);

		
		NetworkConnectionBase conn = null; 
		int deficit=0; 
		int random_token=0;
		int total_size = 0; 
		    

		//if seed mode, then round-robin 
		if (seed_mode) {
		    while(num_ready_connections > 0) { 
			if (cur_conn_index > num_ready_connections -1) {
			    cur_conn_index = 0; 
			} 
			NetworkConnectionBase tmp_conn = (NetworkConnectionBase)ready_array[cur_conn_index];
			cur_conn_index++; 
			
			InetSocketAddress notional_address = tmp_conn.getEndpoint().getNotionalAddress();
			String ip    = notional_address.getAddress().getHostAddress();
			
			int dmessages = tmp_conn.getOutgoingMessageQueue().alexGetNumDataMsgs();
			int first_bytes = tmp_conn.getOutgoingMessageQueue().alexGetBytesFirstMsg();
			
			//ALogger.log(-1, "FT SELECT " + ip + " (D,F) (" + dmessages + "," + first_bytes + ")");
			

			if (!tmp_conn.getTransportBase().isReadyForWrite(waiter)) {
			    num_ready_connections--;
			    continue;
			} 
			
			int tmp_total_size = tmp_conn.getOutgoingMessageQueue().getTotalSize();
			if (tmp_total_size < 1) {
			    num_ready_connections --;
			    continue; //no messages to send on this connection
			} 
			
			conn = tmp_conn;
			total_size = tmp_total_size;

			break; 

		    }//while(1) 

		} 
		
		else { 
		    // ... leecher 
		    
		    //ALEX ft stack optimization where we pop off the last
		    //connection with negative deficit off the stack 
		    if ( AManager.getConfObject().ft_stack_opt) {
			while (true) { 
			    aPeerConn apc =  AManager.getStatsObject().popConnFromFtOptStack();
			    
			    if (apc == null) {
				break; 
			    }
			    
			    NetworkConnectionBase tmp_conn = (NetworkConnectionBase)apc.conn;
			    //check that ready_connections contains our conn
			    //and the peer id of the conn is as expected 
			    if (!ready_connections.contains(tmp_conn) ||
				
				tmp_conn.getAlexPeerId() != apc.peer_id || 
				!tmp_conn.getTransportBase().isReadyForWrite(waiter) ||
				tmp_conn.getOutgoingMessageQueue().getTotalSize() < 1) { 
				
				continue;
			    } 
			    
			    ALogger.log(ALogger.LOG_DEBUG, "FTSTACK can use: " + apc.ip);
			    conn = tmp_conn;
			    break;
			}
		    }
			

		    if (conn == null) { 
			for (Iterator i = ready_connections.iterator(); i.hasNext();) {
			    
			    NetworkConnectionBase tmp_conn = (NetworkConnectionBase)i.next();
			    
			    
			    byte [] peer_id = tmp_conn.getAlexPeerId();
			    
			    
			    
			    int tmp_deficit = 0; 
			    int tmp_random_token  = 1000;
			    
			    
			    if (peer_id != null) {
				tmp_deficit =   AManager.getStatsObject().getDeficit(peer_id);
				tmp_random_token = AManager.getStatsObject().getRandomToken(peer_id);
			    } 
			    
			    InetSocketAddress notional_address = tmp_conn.getEndpoint().getNotionalAddress();
			    String ip    = notional_address.getAddress().getHostAddress();
			    
			    
			    int dmessages = tmp_conn.getOutgoingMessageQueue().alexGetNumDataMsgs();
			    int first_bytes = tmp_conn.getOutgoingMessageQueue().alexGetBytesFirstMsg();
			    
			    
			    int writable = 1; 
			    if (!tmp_conn.getTransportBase().isReadyForWrite(waiter)) {
				writable = 0; 
			    } 
			    
			    int tmp_total_size = tmp_conn.getOutgoingMessageQueue().getTotalSize();
			    
			    
			    //ALogger.log(-1, "FT SELECT " + ip + " def= " + tmp_deficit + " rand = " + tmp_random_token + " (W,T) (" + writable + "," + tmp_total_size + ") (D,F) (" + dmessages + "," + first_bytes + ")");
			    if (writable==0 || tmp_total_size < 1) {
				num_ready_connections --;
				continue; //not writable or no data to send on conn 
			    } 

			    if (conn == null || 
				tmp_deficit < deficit || 
				(tmp_deficit == deficit && tmp_random_token < random_token)) {  
				
				conn = tmp_conn; 
				deficit = tmp_deficit; 
				random_token = tmp_random_token; 
				
				total_size = tmp_total_size; 
			    } 
			    
			}//for...
		    }// if (conn == null)
		    
		}//else ... leecher case 
		
		if (conn == null) {
		    break; 
		} 
		
		
		InetSocketAddress notional_address = conn.getEndpoint().getNotionalAddress();
		String ip    = notional_address.getAddress().getHostAddress();
		//ALogger.log(0, "FT SELECTED " + ip + " def " + deficit + " rand " + random_token);

		int mss_size = conn.getMssSize();
		int use_size = mss_size; 
		if (AManager.getConfObject().fairtorrent_msg_size > 0) {
		    use_size = AManager.getConfObject().fairtorrent_msg_size;
		} 

		//int num_bytes_allowed = num_bytes_remaining > use_size ? use_size : num_bytes_remaining;  //allow a single full packet at most
		//int num_bytes_available = total_size > use_size ? use_size : total_size;  //allow a single full packet at most
		//changing above logic: send all of remaining msg_bytes or the "use_size"  whichever is smaller 
		//if use_size = 17000 we should always be sending a full 16KB data message 
		int msg_bytes = conn.getOutgoingMessageQueue().alexGetBytesFirstMsg();
		int num_bytes_available = use_size > msg_bytes ? msg_bytes : use_size; 

	       
		
		//ALEX
		//ALogger.log(0, "UPL: mss_size: " + mss_size + " allowed: " + num_bytes_allowed + " available: " + num_bytes_available); 
		if(true) {
		    int written = 0;
		    try {
			
			//conn.getOutgoingMessageQueue().alex_last_data_written = 0; 
			//alex: by passing "false" we are causing the notifications to be sent immediately as the data is written
			written = conn.getOutgoingMessageQueue().deliverToTransport( num_bytes_available, false);
			//int data_written = conn.getOutgoingMessageQueue().alex_last_data_written;

			//alex: because we passed false above, notifications have already been sent 
			//if( written > 0 ) {  
			//    manual_notifications.add( conn );  //register it for manual listener notification
			//}
			
			//alex: we don't do waiting list optimization so comment this stuff: 
			//boolean has_urgent_data = conn.getOutgoingMessageQueue().hasUrgentMessage();
			//int remaining = conn.getOutgoingMessageQueue().getTotalSize();
			
			//if( remaining >= mss_size || has_urgent_data ) {  //still has a full packet's worth, or has urgent data
			//    ready_connections.addLast( conn );  //re-add to end for further writing
			//    num_unusable_connections = 0;  //reset the unusable count so that it has a chance to try this connection again in the loop
			//}
			//else {  //connection does not have enough for a full packet, so remove and place into waiting list
			    
			//    addToWaitingList( conn );
			//}
		    }
		    catch( Throwable e ) {  //write exception, so move to waiting list while it waits for removal
			
			if( AEDiagnostics.TRACE_CONNECTION_DROPS ) {
			    if( e.getMessage() == null ) {
				Debug.out( "null write exception message: ", e );
			    }
			    else {
				if( e.getMessage().indexOf( "An existing connection was forcibly closed by the remote host" ) == -1 &&
				    e.getMessage().indexOf( "Connection reset by peer" ) == -1 &&
				    e.getMessage().indexOf( "Broken pipe" ) == -1 &&
				    e.getMessage().indexOf( "An established connection was aborted by the software in your host machine" ) == -1 ) {
				    
				    System.out.println( "MP: write exception [" +conn.getTransportBase().getDescription()+ "]: " +e.getMessage() );
				}
			    }
			}
			
			if (! (e instanceof IOException )){
			    
			    Debug.printStackTrace(e);
			}
			
			connections_to_notify_of_exception.put( conn, e );  //do exception notification outside of sync'd block
			//alex: not waiting list necessary 
			//addToWaitingList( conn );
		    }
		    //ALogger.log(0, "FT SELECTED " + ip + " attempted " + num_bytes_available + " wrote " + written + " bytes ");
		    if (written > 0) { 
			num_bytes_remaining -= written;
		    }
		    else {
			//ALogger.log(0, "REMOVING " + ip + " ready_connections: " + ready_connections.size());
			leftover_conns.add(conn);
			num_ready_connections--;
			ready_connections.remove(conn);
		    } 

		}
		//else {  //we're not allowed enough to maximize the packet payload
		//    ready_connections.addLast( conn );  //re-add to end as currently unusable
		//    num_unusable_connections++;
		    
		//    continue;  //move on to the next connection
		//}
	    }

	    
	    

	}
	finally {
	    lists_lock.exit();
	}
	
	//manual queue listener notifications
	for( int i=0; i < manual_notifications.size(); i++ ) {
	    NetworkConnectionBase conn = (NetworkConnectionBase)manual_notifications.get( i );
	    conn.getOutgoingMessageQueue().doListenerNotifications();
	}
	
	
	//exception notifications
	for( Iterator i = connections_to_notify_of_exception.entrySet().iterator(); i.hasNext(); ) {
	    Map.Entry entry = (Map.Entry)i.next();
	    NetworkConnectionBase conn = (NetworkConnectionBase)entry.getKey();
	    Throwable exception = (Throwable)entry.getValue();
	    conn.notifyOfException( exception );
	}

	//add leftover conns back to the ready conns 
	for( int i=0; i < leftover_conns.size(); i++ ) {
	    NetworkConnectionBase conn_l = (NetworkConnectionBase)leftover_conns.get( i );
	    ready_connections.addLast(conn_l);
	}
	
	int num_bytes_written = num_bytes_to_write - num_bytes_remaining;
	ALogger.log(ALogger.LOG_DEBUG, "FT written: " + num_bytes_written);
	//rate_handler.bytesProcessed( num_bytes_to_write);
	//return  num_bytes_to_write;
	
	if( num_bytes_written > 0 ) {
	    rate_handler.bytesProcessed( num_bytes_written );
	}

	last_bytes_written = num_bytes_written; 
	if (last_bytes_written > 0) {
	    last_time_written = SystemTime.getCurrentTime();
	} 

	return num_bytes_written;
       
    }

  private int write( EventWaiter waiter, int num_bytes_to_write ) {  //TODO: model this class after the simplicity of MultiPeerDownloader
    if( num_bytes_to_write < 1 ) {
      Debug.out( "num_bytes_to_write < 1" );
      return 0;  //not allowed to write
    }

    //ALEX call fairtorrent_write
    if (AManager.getConfObject().fairtorrent_mode && !AManager.getConfObject().done_with_download) {
	return fairtorrent_write(waiter, num_bytes_to_write, false);
    }
    else if (AManager.getConfObject().fairtorrent_mode && 
	     AManager.getConfObject().done_with_download && 
	     AManager.getConfObject().seed_fairtorrent_mode) {
	//seed round-robin mode 
	return fairtorrent_write(waiter, num_bytes_to_write, true);
    }
        
    HashMap connections_to_notify_of_exception = new HashMap();
    ArrayList manual_notifications = new ArrayList();
    
    int num_bytes_remaining = num_bytes_to_write;    
    
    try {
      lists_lock.enter();
      
      int num_unusable_connections = 0;
            
      while( num_bytes_remaining > 0 && num_unusable_connections < ready_connections.size() ) {
    	NetworkConnectionBase conn = (NetworkConnectionBase)ready_connections.removeFirst();
        
        if( !conn.getTransportBase().isReadyForWrite( waiter ) ) {  //not yet ready for writing
          ready_connections.addLast( conn );  //re-add to end as currently unusable
          num_unusable_connections++;
          continue;  //move on to the next connection
        }
        
        int total_size = conn.getOutgoingMessageQueue().getTotalSize();
        
        if( total_size < 1 ) {  //oops, all messages have been removed
          addToWaitingList( conn );
          continue;  //move on to the next connection
        }
        
        int mss_size = conn.getMssSize();
        int num_bytes_allowed = num_bytes_remaining > mss_size ? mss_size : num_bytes_remaining;  //allow a single full packet at most
        int num_bytes_available = total_size > mss_size ? mss_size : total_size;  //allow a single full packet at most

	//ALEX
	//ALogger.log(0, "UPL: mss_size: " + mss_size + " allowed: " + num_bytes_allowed + " available: " + num_bytes_available); 
        if( num_bytes_allowed >= num_bytes_available ) { //we're allowed enough (for either a full packet or to drain any remaining data)
          int written = 0;
          try {
            written = conn.getOutgoingMessageQueue().deliverToTransport( num_bytes_available, true );
                   
            if( written > 0 ) {  
              manual_notifications.add( conn );  //register it for manual listener notification
            }
            
            boolean has_urgent_data = conn.getOutgoingMessageQueue().hasUrgentMessage();
            int remaining = conn.getOutgoingMessageQueue().getTotalSize();
            
            if( remaining >= mss_size || has_urgent_data ) {  //still has a full packet's worth, or has urgent data
              ready_connections.addLast( conn );  //re-add to end for further writing
              num_unusable_connections = 0;  //reset the unusable count so that it has a chance to try this connection again in the loop
            }
            else {  //connection does not have enough for a full packet, so remove and place into waiting list
 
              addToWaitingList( conn );
            }
          }
          catch( Throwable e ) {  //write exception, so move to waiting list while it waits for removal
            
            if( AEDiagnostics.TRACE_CONNECTION_DROPS ) {
              if( e.getMessage() == null ) {
                Debug.out( "null write exception message: ", e );
              }
              else {
                if( e.getMessage().indexOf( "An existing connection was forcibly closed by the remote host" ) == -1 &&
                    e.getMessage().indexOf( "Connection reset by peer" ) == -1 &&
                    e.getMessage().indexOf( "Broken pipe" ) == -1 &&
                    e.getMessage().indexOf( "An established connection was aborted by the software in your host machine" ) == -1 ) {
                  
                  System.out.println( "MP: write exception [" +conn.getTransportBase().getDescription()+ "]: " +e.getMessage() );
                }
              }
            }
            
            if (! (e instanceof IOException )){
            	
          	  Debug.printStackTrace(e);
            }
            
            connections_to_notify_of_exception.put( conn, e );  //do exception notification outside of sync'd block
            addToWaitingList( conn );
          }
          
          num_bytes_remaining -= written;
        }
        else {  //we're not allowed enough to maximize the packet payload
          ready_connections.addLast( conn );  //re-add to end as currently unusable
          num_unusable_connections++;
          
          continue;  //move on to the next connection
        }
      }
    }
    finally {
      lists_lock.exit();
    }
    
    //manual queue listener notifications
    for( int i=0; i < manual_notifications.size(); i++ ) {
      NetworkConnectionBase conn = (NetworkConnectionBase)manual_notifications.get( i );
      conn.getOutgoingMessageQueue().doListenerNotifications();
    }
    
    //exception notifications
    for( Iterator i = connections_to_notify_of_exception.entrySet().iterator(); i.hasNext(); ) {
      Map.Entry entry = (Map.Entry)i.next();
      NetworkConnectionBase conn = (NetworkConnectionBase)entry.getKey();
      Throwable exception = (Throwable)entry.getValue();
      conn.notifyOfException( exception );
    }
    
    int num_bytes_written = num_bytes_to_write - num_bytes_remaining;
    if( num_bytes_written > 0 ) {
      rate_handler.bytesProcessed( num_bytes_written );
    }
    
    return num_bytes_written;
  }
  
  
  
  /**
   * Does this entity have data ready for writing.
   * @return true if it has data to send, false if empty
   */
  /*
  public boolean hasWriteDataAvailable() {
    if( ready_connections.isEmpty() )  return false;
    return true;
  }
  */
  
  
  
  private static class PeerData {
    private OutgoingMessageQueue.MessageQueueListener queue_listener;
    private long last_message_added_time;
    private boolean	bumped;
  }
  
  public long
  getBytesReadyToWrite()
  {
	  long	total = 0;
	  
	  try {
		  lists_lock.enter();

		  for( Iterator i = waiting_connections.keySet().iterator(); i.hasNext(); ) {
			 
			  NetworkConnectionBase conn = (NetworkConnectionBase)i.next();
			  
			  total += conn.getOutgoingMessageQueue().getTotalSize();
		  }
		  
		  for( Iterator i = ready_connections.iterator(); i.hasNext(); ) {
				 
			  NetworkConnectionBase conn = (NetworkConnectionBase)i.next();
			  
			  total += conn.getOutgoingMessageQueue().getTotalSize();
		  }
	  }finally{
		  
		  lists_lock.exit();
	  }
	  
	  return( total );
  }
  
  public int
  getConnectionCount()
  {
	  return( waiting_connections.size() + ready_connections.size());
  }
  
  public int
  getReadyConnectionCount(
	EventWaiter	waiter )
  {
	  int	total = 0;
	  
	  try {
		  lists_lock.enter();

		  for( Iterator i = waiting_connections.keySet().iterator(); i.hasNext(); ) {
			 
			  NetworkConnectionBase conn = (NetworkConnectionBase)i.next();
			  
			  if ( conn.getTransportBase().isReadyForWrite(waiter)){
				  
				  total++;
			  }
		  }
		  
		  for( Iterator i = ready_connections.iterator(); i.hasNext(); ) {
				 
			  NetworkConnectionBase conn = (NetworkConnectionBase)i.next();
			  
			  if ( conn.getTransportBase().isReadyForWrite(waiter)){
				  
				  total++;
			  }
		  }
	  }finally{
		  
		  lists_lock.exit();
	  }
	  
	  return( total );
  }
  
  //////////////// RateControlledWriteEntity implementation ////////////////////
  
  public boolean canProcess( EventWaiter waiter ) {
    flushCheck();  //since this method is called repeatedly from a loop, we can use it to check flushes

    if( ready_connections.isEmpty() )  return false;  //no data to send

    //ALEX_DEBUG
    long now = SystemTime.getCurrentTime();

    //ALEX: don't print this for now... 
    if (false) { 
	int ready_num = ready_connections.size();
	
	
	for (Iterator i = ready_connections.iterator(); i.hasNext();) { 
	    NetworkConnectionBase connection = (NetworkConnectionBase)i.next();
	    
	    int alex_ts = connection.getOutgoingMessageQueue().getTotalSize();
	    int alex_av = rate_handler.getCurrentNumBytesAllowed();
	    if (alex_ts >= 1 && alex_av <1 && 
		now - last_check > 2000) {
		last_check = now; 
		
		InetSocketAddress notional_address = connection.getEndpoint().getNotionalAddress();
		
		String ip    = notional_address.getAddress().getHostAddress();
		
		int npriority = AManager.getStatsObject().getNumHighPriorityConnections();
		
		String my_client = AManager.getConfObject().getMyClientType();
		
		byte [] peer_id = connection.getAlexPeerId();
		
		int am_seed = 0; 
		if (peer_id != null) {
		    apeer ap = AManager.getStatsObject().getPeerById(peer_id);
		    if (ap != null && ap.am_seed) { 
			am_seed = 1;
		    } 
		} 
		
		ALogger.log(ALogger.LOG_DEBUG2, "BTYDEBUG WRITE_NOT_PROCESSED " + ip + " bytes to write " + alex_ts + " S:" + am_seed + " HIGHP CONNS: " + npriority + " READY: " + ready_num + " client: " + my_client);
	    }
	}
    } 
    
    if( rate_handler.getCurrentNumBytesAllowed() < 1/*NetworkManager.getTcpMssSize()*/ )  return false;
    return true;
  }
  
  public boolean doProcessing( EventWaiter waiter ) {
    int num_bytes_allowed = rate_handler.getCurrentNumBytesAllowed();
    if( num_bytes_allowed < 1 )  return false;
    
    return write( waiter, num_bytes_allowed ) > 0 ? true : false;
  }

  public int getPriority() {
    return RateControlledEntity.PRIORITY_HIGH;
  }
  
  public String
  getString()
  {
	  String	str = "can_process=" + canProcess(null) + ",bytes_allowed=" + rate_handler.getCurrentNumBytesAllowed() + ", waiting ";
  
	  try {
		  lists_lock.enter();

		  for( Iterator i = waiting_connections.keySet().iterator(); i.hasNext(); ) {
			 
			  NetworkConnectionBase conn = (NetworkConnectionBase)i.next();
			  
			  str += "," + conn.getString();
		  }
		  
		  str += ": ready ";
		  
		  for( Iterator i = ready_connections.iterator(); i.hasNext(); ) {
				 
			  NetworkConnectionBase conn = (NetworkConnectionBase)i.next();
			  
			  str += "," + conn.getString();
		  }
	  }finally{
		  
		  lists_lock.exit();
	  }
	  
	  return( "MPU: " + str );
  }  
}
